# coding:utf-8
# *******************************************************************
# * Copyright 2023-present evilbinary
# * 作者: evilbinary on 01/01/20
# * 邮箱: rootdebug@163.com
# ********************************************************************
import os
import math
import platform 
import sys

module_path=['..',
    '.',
    '../../',
    '../../..',
    'xenv']
sys.path.extend(module_path)

import json
from colorama import Fore, Back, Style, init
import fnmatch
import types
import argparse
import inspect
import hashlib
import datetime
import logging
from log import log
from function import *
from builder import *
from toolchain import toolchains_init
from version import version
from globa import verborse,jobnum,mode,is_init,is_process
from graph import build_graph,find_cycles

true=True
false=False
args=None
parser=None


def compile(project,graph,name):
    G = nx.DiGraph(graph)
    # 执行拓扑排序
    topological_order = [] 

    toolchain_name=project.get('toolchain')

    if name:
        t=project.get('target-objs').get(name)
        if not t:
            log.error('not found target {}'.format(name))
            return
        topological_order=get_dep_order(t)
        topological_order.reverse()
        topological_order.append(name)
    else:
        topological_order=(list(nx.topological_sort(G)))
        topological_order.reverse()
   
    total_nodes=len(topological_order)

    # 根据拓扑排序的顺序反向遍历节点，并编译代码
    compiled_code = []
    start_time = datetime.datetime.now()
    build_target=[]
    for i,node in enumerate(topological_order):
        # 编译节点对应的代码
        # print('node===>',node,project.get('target-objs'))

        target=project.get('target-objs').get(node)

        if not target:
            log.error('not found target {}'.format(node))
            return
        if target.get('toolchain'):
            toolchain_name=target.get('toolchain')
        build_target.append(target)

        toolchain=nodes_get_type_and_name('toolchain',toolchain_name)
        if not toolchain:
            log.error('not found toolchain {}'.format(toolchain_name))
            break

        opt={
            'progress':i,
            'total_nodes':total_nodes
        }
        bp=toolchain.get('build_prepare')
        if bp:
            bp(toolchain,target,opt)
        toolchain.get('build')(toolchain,target,opt)
        progress = i + 1
        print_progress('compile',progress,total_nodes,node)


    end_time = datetime.datetime.now()
    time_diff=end_time-start_time
    if len(build_target)>0:
        print("{}build success!{} cost {}".format(Fore.GREEN,Style.RESET_ALL,time_diff) )
    else:
        print("{}build nothing not target found!{} cost {}".format(Fore.GREEN,Style.RESET_ALL,time_diff) )

class CustomEncoder(json.JSONEncoder):
    cache={}
    def default(self, obj):
        try:        
            if callable(obj) or isinstance(obj, type):
                return str(obj)
            if not cache.get(obj):
                return super().default(obj)
                cache[obj]=True
        except TypeError:
            pass

def node_dump(node):
    if not node:
        return
    d=json.dumps(node,cls=CustomEncoder,indent=4)
    print(node.get('type'),d )
    pass


def build(name=None):
    node_finish()
    for p in nodes:
        if not p.get('type') in ['project']:
            continue
        # print('project=>',p)
        graph=build_graph(p)
        # print('graph=>',graph)

        if len(graph)<=0:
            print('{}build nothing finish.{}'.format(Fore.GREEN,Style.RESET_ALL))
            continue
        cycles=find_cycles(graph)
        
        if len(cycles)>0:
            print("cycle depency:")
            for cycle in cycles:
                print(' -> '.join(cycle))
            print('build failed')
        else:
            compile(p,graph,name)



def clean_target(project,graph,name):
    G = nx.DiGraph(graph)
    # 执行拓扑排序
    topological_order = list(nx.topological_sort(G))
    total_nodes = len(topological_order)
    toolchain_name=project.get('toolchain')

    # 根据拓扑排序的顺序反向遍历节点，并编译代码
    compiled_code = []
    start_time = datetime.datetime.now()
    build_target=[]
    for i,node in enumerate(reversed(topological_order)):
        # 编译节点对应的代码
        # print('node===>',node,project.get('target-objs'))

        target=project.get('target-objs').get(node)
        if name and target:
            if not (target.get('name')==name ):
                continue
        if not target:
            log.error('not found target {}'.format(node))
            return
        if target.get('toolchain'):
            toolchain_name=target.get('toolchain')
        build_target.append(target)

        toolchain=nodes_get_type_and_name('toolchain',toolchain_name)
        if not toolchain:
            log.error('not found toolchain {}'.format(toolchain_name))
            break

        opt={
            'progress':i,
            'total_nodes':total_nodes
        }
        if toolchain.get('build_prepare'):
            toolchain.get('build_prepare')(toolchain,target,opt)
        print('clean target',toolchain)

        #if toolchain.get('clean'):
        toolchain.get('clean')(toolchain,target,opt)
        progress = i + 1
        print_progress('clean',progress,total_nodes,node)


    end_time = datetime.datetime.now()
    time_diff=end_time-start_time
    if len(build_target)>0:
        print("{}clean success!{} cost {}".format(Fore.GREEN,Style.RESET_ALL,time_diff) )
    else:
        print("{}clean nothing not target found!{} cost {}".format(Fore.GREEN,Style.RESET_ALL,time_diff) )

def clean(name=None):
    node_finish()
    for p in nodes:
        if not p.get('type') in ['project']:
            continue
        # print('project=>',p)
        graph=build_graph(p)
        # print('graph=>',graph)

        if len(graph)<=0:
            print('{}clean nothing finish.{}'.format(Fore.GREEN,Style.RESET_ALL))
            continue
        cycles=find_cycles(graph)
        
        if len(cycles)>0:
            print("cycle depency:")
            for cycle in cycles:
                print(' -> '.join(cycle))
            print('clean failed')
        else:
            clean_target(p,graph,name)


def run(name):
    node_finish()

    target=nodes_get_type_and_name('target',name)
    if target:
        call_hook_event(target,'before_run')
        call_hook_event(target,'on_run')
    else:
        print('target {}{}{} not found to run, target list: '.format(Fore.MAGENTA,name,Style.RESET_ALL))

        targets=nodes_get_all_type('target')
        for target in targets:
            print('    {}'.format(target.get('name')))
    
def load():
    ya=os.path.join(os.getcwd(),"./ya.py")
    ya=os.path.normpath(ya)
    add_subs(ya)

def process_option(parser):
    options=nodes_get_all_type('option')
    for o in options:
        if o.get('args'):
            parser.add_argument('--'+o.get('name'),nargs=o.get('args'), default=o.get('default'), help=o.get('description'))
        else:
            parser.add_argument('--'+o.get('name'),action='store_true', default=o.get('default'), help=o.get('description'))
    return options

def init():
    global is_init,parser,args,verborse
    root('root')
    add_toolchain_dirs('toolchains')
    
    # 默认工具
    toolchain('gcc',build=gcc_build,build_prepare=build_prepare,clean=gcc_clean)
    if platform.system()=='Darwin':
        add_ldflags('-lSystem')
        #add_ldflags('-arch', platform.machine())
    elif platform.system()=='Linux':
        add_ldflags('-lc')
        cur=node_current()
        prefix= cur.get('prefix')
        set_toolset('ld',prefix+'gcc')

    toolchain('arm-none-eabi',
        prefix='arm-none-eabi-',
        build=gcc_build,
        build_prepare=build_prepare,
        clean=gcc_clean)
    toolchain('riscv64-unknown-elf',
        prefix='riscv64-unknown-elf-',
        build=gcc_build,
        build_prepare=build_prepare,
        clean=gcc_clean)
    toolchain('i386-elf',
        prefix='i386-elf-',
        build=gcc_build,
        build_prepare=build_prepare,
        clean=gcc_clean)
    toolchain('i686-elf',
        prefix='i686-elf-',
        build=gcc_build,
        build_prepare=build_prepare,
        clean=gcc_clean)

    toolchain('gcc-13',
    prefix='gcc-',
    suffix='-13',
    build=gcc_build,build_prepare=build_prepare)
    add_ldflags('-lc')
    cur=node_current()
    prefix= cur.get('prefix')
    set_toolset('ld',prefix+'')
    set_toolset('cc','gcc-13')

    toolchain('aarch64-none-elf',
        prefix='aarch64-none-elf-',
        build=gcc_build,
        build_prepare=build_prepare,
        clean=gcc_clean)
    
    toolchain('xtensa-esp32-elf',
    prefix='xtensa-esp32-elf-',
    build=gcc_build,
    build_prepare=build_prepare,
    clean=gcc_clean)

    toolchains_init()
    toolchain_end()


    rule('mode.debug')
    def build_config(target):
        if is_mode("debug"):
            target.add('cflags',"-g")
        pass
    on_config(build_config)
    rule_end()

    rule('mode.release')
    def build_config(target):
        if is_mode("release"):
            target.add('cflags',"-O2")
        pass
    on_config(build_config)
    rule_end()

    print('welcome to use {}ymake{} {} ,make world happy ^_^!!'.format(Fore.GREEN,Style.RESET_ALL,version))

    # 创建参数解析器
    parser = argparse.ArgumentParser(allow_abbrev=True,add_help=False)

    parser.add_argument('-v',nargs='?', default=None, help='verborse info debug error')
    parser.add_argument('-r','-run',nargs='?', default=None, help='run the project target.')
    parser.add_argument('-j',nargs='?', default=1, help='job number')
    parser.add_argument('-m','-mode',nargs='?', default=None, help='build mode debug release')
    parser.add_argument('-b','-build',nargs='?', default=None, help='build the project target.')
    parser.add_argument('-c','-clean',nargs='?', default=None, help='clean the project target.')
    parser.add_argument('-p','-plat',nargs='?', default=None, help='select the project platform.')

    parser.add_argument('-h','--help',action='store_true', default=None, help='help')

    args,unknown = parser.parse_known_args()

    verborse=args.v
    if args.help :
        load()
        process_option(parser)
        # 解析命令行参数
        args = parser.parse_args()
        parser.print_help()

        exit(0)
   
    log.debug('args=>',unknown)
    # 添加参数
    for name in unknown:
        n=name.replace('--','')
        option(n,value=True)        

    if args.v=='D':
        log.setLevel(logging.DEBUG)
    elif args.v=='I':
        log.setLevel(logging.INFO)
    elif args.v=='W':
        log.setLevel(logging.WARN)
    if args.j:
        jobnum=args.j
        set_config('jobnum',int(jobnum))
    if args.m:
        mode=args.m
        set_config('mode',mode)
    if args.p:
        set_defaultplat(args.p)

def process():
    try:
        global mode,jobnum,is_process
        if is_process:
            return

        is_process=True
       
        if args.c:
            clean(args.c)
        elif args.b:
            build(args.b)
        else:
            build(args.b)
        if args.r:
            run(args.r)
    except (KeyboardInterrupt):
        pass
    except Exception as e:
        if verborse=='D':
            raise e
        pass

if __name__ == 'yaya':
    init()
    load()
    process()

else:
    pass



# coding:utf-8
# *******************************************************************
# * Copyright 2023-present evilbinary
# * 作者: evilbinary on 01/01/20
# * 邮箱: rootdebug@163.com
# ********************************************************************
from function import toolchain,add_buildin,node_start,set_toolset,set_kind

def build(tool,target,opt={}):
    print('cmake build',tool)
    pass
    
def cmake(name, **kwargs):
    node={
        'name': name,
        'type':'cmake',
        'toolchain':'cmake'
    }
    node.update(kwargs)
    node_start(node)

def cmake_end():
    cur=node_current()
    if cur and cur.get('type')=='cmake':
        node_end()

def init():
    toolchain('cmake',build=build)
    set_kind("standalone")
    #set toolset
    set_toolset("cmake","cmake")

    add_buildin('cmake',cmake)


# coding:utf-8
# *******************************************************************
# * Copyright 2023-present evilbinary
# * 作者: evilbinary on 01/01/20
# * 邮箱: rootdebug@163.com
# ********************************************************************
import os
from function import toolchain,add_buildin,node_start,set_toolset,\
    set_kind,node_current,shell,node_set,cmd,get_config,node_get_parent,\
    node_get_formated
from builder import get_build_target
from log import log
from node import Node

def build_prepare(tool,target,opt={}):
    build_dir=node_get_formated(target,'build-dir')
    build_dir_abs=os.path.abspath(build_dir)

    build_dir_target=os.path.join(build_dir_abs,target.get('name'))
    build_dir_target_re= os.path.join(build_dir,target.get('name'))

    automake=target.get('build-tool')

    target['build-lib-dir']=build_dir_target_re
    target['build-dir']=build_dir_target_re+'/lib'
    automake['build-dir']=build_dir_target_re+'/lib'



def build(tool,target,opt={}):
    log.debug('{} build {}'.format(tool.get('name'),target.get('name')))

    automake=target.get('build-tool')
    sourcedir=target.get('sourcedir')
    args=automake.get('configure')
    jobnum= node_get_parent(automake,'jobnum')
    build_dir=node_get_formated(target,'build-lib-dir')

    build_dir_abs=os.path.abspath(build_dir)
    build_config_abs=os.path.abspath(os.path.join(sourcedir,tool.get('configure')))



    build_target=get_build_target(target)


    build_target_abs=os.path.abspath(build_target)


    log.debug('build_config_abs=>{}'.format(build_config_abs))
    log.debug('build_target_abs=>{}'.format(build_target_abs))

    is_modify_target=False
    if os.path.exists(build_target_abs):
        is_modify_target=True
        return

    if not os.path.exists(build_config_abs):
        shell('aclocal')
        shell('autoconf')
        shell('automake --add-missing')

    args+=['--prefix='+build_dir_abs]
    try:
        shell(tool.get('configure'),args,cwd=sourcedir)
        
    except Exception as e:
        print('build error',e)
        pass
    try:
        shell(tool.get('make'),['install','-j'+str(jobnum)],cwd=sourcedir)
    except Exception as e:
        shell(tool.get('make'),['-j'+str(jobnum)],cwd=sourcedir)
        pass

    pass
    
def automake(name=None, **kwargs):
    cur=node_current()
    if name==None:
        name=cur.get('name')

    data={
        'name': name,
        'type':'automake',
        'toolchain':'automake'
    }
    n=Node(**data)
    n.update(kwargs)
    if cur:
        cur['build-tool']=n    
    node_start(n)

def automake_end():
    cur=node_current()
    if cur and cur.get('type')=='automake':
        node_end()

def configure(*args):
    node_set('configure',list(args))


def init():
    toolchain('automake',build=build,build_prepare=build_prepare)
    set_kind("standalone")

    set_toolset("configure","./configure")
    set_toolset("automake","automake")
    set_toolset("autoconf","autoconf")
    set_toolset("libtool","libtool")
    set_toolset("pkg-config","pkg-config")
    set_toolset("make","make")

    add_buildin('automake',automake)
    add_buildin('configure',configure)




# coding:utf-8
# *******************************************************************
# * Copyright 2023-present evilbinary
# * 作者: evilbinary on 01/01/20
# * 邮箱: rootdebug@163.com
# ********************************************************************
from log import log
import os
import inspect

nodes=[]

node_stack=[]


node_level={
    'project':0,
    'target':1,
    'rule':1,
    'toolchain':1,
    'option':1
}

def node_set_level(k,v):
    node_level[k]=v

def get_format(str,data):
    try:
        return str.format(**data)
    except Exception as e:
        log.warn('get format {} {} {}'.format(str,e,data))
        return str

def get_formats(l,data):
    ret=[]
    for i in l:
        new=get_format(i,data)
        ret.append(new)
    return ret

def get_target_data(target):
    data={}
    target_name=node_get_parent(target,'name')
    target_plat=node_get_parent(target,'plat')
    target_mode=node_get_parent(target,'mode')
    target_arch=node_get_parent(target,'arch')
    target_build_dir=node_get_parent(target,'build-dir')

    data['name']=target_name
    data['plat']=target_plat
    data['mode']=target_mode
    data['arch']=target_arch

    target_build_dir=get_format(target_build_dir,data)

    data['buildir']=target_build_dir
    return data

def format_target_var(target,var):
    data=get_target_data(target)
    return get_format(var,data)


def node_get_formated(target,key):
    build_dir=target.get(key)
    if not build_dir:
        if target.get('project'):
            build_dir=target.get('project').get(key)

    if not build_dir:
        return None
    build_dir=format_target_var(target,build_dir)
    # print(target.get('name'),'build dir=',build_dir)

    build_dir=os.path.normpath(build_dir)

    return build_dir


def get_list_args(args,sep=' '):
    ret=[]
    for a in args:
        if isinstance(a,str):
            s=a.split(sep)
            if len(s)>0:
                ret+=[x for x in s if x != '']
            else:
                ret.append(a.strip())
        else:
            ret+=a
    return ret

class Node(dict):
    
    # obj={}
    # def __init__(self,val,**vals):
    #     self.obj.update(val)
    #     super(val)

    # def update(self,vals):
    #     self.obj.update(vals)
    def name(self):
        return self.get('name')

    def targetfile(self):
        build_dir=node_get_formated(self,'build-dir')
        name=self.get('name')
        if self.get('filename'):
            name= self.get('filename')
        if build_dir:
            return build_dir+'/'+name
        else:
            return None

    def sourcefiles(self):
        files=self.get('files')
        ret=[]
        for file in files:
            file=format_target_var(self,file)
            ret.append(file)
        return ret

    def objectfiles(self):
        return self.get('file-objs')

    def set(self,key,val):
        self[key]=val

    def toolchain(self):
        toolchain_name=  self.get('toolchain')
        if not toolchain_name:
            toolchain_name = node_get_parent(self,'toolchain')
        toolchain=nodes_get_type_and_name('toolchain',toolchain_name)
        
        return toolchain

    def tool(self,tool):
        toolchain_name= node_get_parent(self,'toolchain')

        toolchain=nodes_get_type_and_name('toolchain',toolchain_name)
        if toolchain:
            return toolchain.get(tool)
        return None

    def get(self,key):
        if key in['arch_type','arch','plat' ]:
            return node_get_parent(self,key)
        if key in self:
            return self[key]
        return None

    def get_arch(self):
        return self.get('arch')

    def arch(self):
        return self.get('arch')

    def arch_type(self):
        return self.get('arch_type')

    def get_arch_type(self):
        return self.get('arch_type')

    def add(self,key,*val,**kwargs):
        after=kwargs.pop('after',False)
        vals=get_list_args(val)
  
        node_op_extend(self,key,vals,after,True)

    def plat(self):
        return node_get_parent(self,'plat')

    def get_config(self,key):
        return node_get_parent(self,key)

def is_same_level(n1,n2):
    if node_level.get(n1.get('type'))== node_level.get(n2.get('type')):
        return True
    return False

def node_start(n):
    cur=node_current()
    if cur:
        # print('n type',n.get('type'),'====',cur.get('type'),'=>',n.get('type')==cur.get('type') )
        if is_same_level(n,cur):
            node_del=node_stack.pop()
            nodes_append(node_del)
        n['parent']=cur
    node_stack.append(n)

    log.debug('{} {}'.format(n.get('type'),n.get('name')))
     
def node_end():
    if len(node_stack)>0:
        n=node_stack.pop()
        nodes_append(n)
        # nodes.append(n)

def node_save():
    n =node_current()
    node_stack.append(n)

def node_restore():
    node_end()

def node_current():
    if len(node_stack)>0:
        return node_stack[-1]
    else:
        return None

def node_set(key,value):
    n=node_current()
    if n:
        n[key]=value
    else:
        log.warn('node current is null')

def node_get(key):
    n=node_current()
    return n.get(key)

def node_get_all(key):
    n =node_current()
    while n:
        if n.get(key):
            return n.get(key)
        n=n.get('parent')
    return None

def node_get_parent(n,key):
    while n:
        if key in n:
            if n[key]:
                return n[key]
        n=n.get('parent')
    return None

def node_get_parent_all(n,key):
    ret=[]
    while n:
        if n.get(key):
            if isinstance( n.get(key),str):
                ret.append(key)
            else:
                # print('+====ret=',ret,'==>',n.get(key))
                ret+=n.get(key)                
        n=n.get('parent')
    return ret

def node_update(data):
    n=node_current()
    return n.update(data)

def node_extend(key,value,t=1):
    n=node_current()
    if not isinstance(value,list):
        value=[value]
    return node_op_extend(n,key,value,t)

def node_op_extend(n,key,value,tail=True,nodup=False):
    if not n.get(key):       
        # caller_frame = inspect.currentframe().f_back
        # caller_file_path = inspect.getframeinfo(caller_frame)
        # print('call==>',caller_file_path)        
        # print('get=>',value,'==>',list(value) )
        if isinstance(value,list):
            # Handle case where list might contain unhashable elements
            try:
                n[key] = list(set(value))
            except TypeError:
                # If elements are unhashable (like lists), keep them as-is without removing duplicates
                n[key] = value.copy() if hasattr(value, 'copy') else list(value)
        elif isinstance(value,tuple):
            n[key]=list(value)
        else:
            n[key]=[value]
        return value
    if nodup:
        # value=list(set(value))
        value= list(sorted(set(value), key=value.index))

    if tail:
        n[key].extend(value)
    else:
        n[key][:0]=value
    if nodup:
        n[key]=list(sorted(set(n[key]), key=n[key].index))
        # n[key]=list(set(n[key]))
    
    # print('get1=>',n.get(key),'==',value)

def node_len():
    return len(nodes)+len(node_stack)


def node_finish():
    global nodes
    # for s in node_stack:
    #     nodes_append(s)
    nodes.extend(node_stack)
    node_stack.clear()
    all_nodes=nodes_merge(nodes)
    nodes.clear()
    nodes+=all_nodes

def nodes_merge(nodes):
    unique_list = []
    seen_names = set()

    for n in nodes:
        name = n.get('name')+n.get('type')
        if name not in seen_names:
            seen_names.add(name)
            unique_list.append(n)
    return unique_list

def nodes_append(n):
    dup=False
    for nn in nodes:
        if n.get('name')==nn.get('name') and n.get('type')== nn.get('type'):
            dup=True
            break
    if not dup:
        nodes.append(n)
    # else:
    #     log.error('dup {}'.format(n.get('name') ))

def nodes_get_type_and_name(ty,name):
    for n in nodes:
        if n.get('type')== ty and n.get('name')==name :
            return n
    for n in node_stack:
        if n.get('type')== ty and n.get('name')==name :
            return n
    return None

def nodes_get_all_type(ty):
    ret=[]
    for n in nodes:
        if n.get('type')== ty:
            ret.append(n)
    for n in node_stack:
        if n.get('type')== ty:
            ret.append(n)
    return ret


from time import time
import setuptools
import os,sys,re

with open("README.md", "r") as fh:
    long_description = fh.read()
setuptools.setup(
    name="yymake",
    version="0.9.1",
    author="evilbinary",
    author_email="rootntsd@gmail.com",
    description="A cross build dsl make tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/evilbinary/ymake",
    project_urls={
        "Bug Tracker": "https://github.com/evilbinary/ymake/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        'console_scripts': [
            'ya = yaya:process',
            'yaya = yaya:process',
            'ymake = yaya:process',
            'yymake = yaya:process',
        ],
    },
    package_dir={"": "ymake"},
    packages=["."],
    install_requires=[
        'pytest',
        'colorama',
        'networkx', 
        'colorlog',
        'diskcache'
    ],
    python_requires=">=2.6",
)